# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################


import sys
import os

from aiida.engine import run
from aiida.plugins.factories import DataFactory
from aiida.orm import Group
from aiida.orm import Code

from wf_aiida_1.0 import TestWorkChain
################################################################

StructureData = DataFactory('structure')

try:
    codename = sys.argv[1]
except IndexError:
    codename = None

if not codename:
    print ("Please provide a valid codename")
    sys.exit(1)
code = Code.get_from_string(codename)

alat = 4.  # angstrom
cell = [[alat, 0., 0., ],
        [0., alat, 0., ],
        [0., 0., alat, ],
]

# BaTiO3 cubic structure
s = StructureData(cell=cell)
s.append_atom(position=(0., 0., 0.), symbols=['Ba'])
s.append_atom(position=(alat / 2., alat / 2., alat / 2.), symbols=['Ti'])
s.append_atom(position=(alat / 2., alat / 2., 0.), symbols=['O'])
s.append_atom(position=(alat / 2., 0., alat / 2.), symbols=['O'])
s.append_atom(position=(0., alat / 2., alat / 2.), symbols=['O'])

g = Group(name="input_group").store()
g.add_nodes(s.store())

w = TestWorkChain
run(w, structure=s.store(), code=code)



# -*- coding: utf-8 -*-
"""Test export file migration from old export versions to the newest"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.backends.tests.utils.configuration import with_temp_dir
from aiida.cmdline.commands.cmd_export import migrate
from aiida.cmdline.utils.migration import migrate_recursively, verify_metadata_version
from aiida.orm.importexport import import_data, EXPORT_VERSION as newest_version

from ..utils import get_archive_file, get_json_files, migrate_archive


class TestExportFileMigration(AiidaTestCase):
    """Test export file migrations"""

    def setUp(self):
        self.reset_database()

        # Known export file content used for checks
        self.node_count = 25
        self.struct_count = 2
        self.known_struct_label = ''
        self.known_cell = [[4, 0, 0], [0, 4, 0], [0, 0, 4]]
        self.known_kinds = [
            {
                'name': 'Ba',
                'mass': 137.327,
                'weights': [1],
                'symbols': ['Ba']
            },
            {
                'name': 'Ti',
                'mass': 47.867,
                'weights': [1],
                'symbols': ['Ti']
            },
            {
                'name': 'O',
                'mass': 15.9994,
                'weights': [1],
                'symbols': ['O']
            },
        ]

    def test_migrate_recursively(self):
        """Test function 'migrate_recursively'"""
        import io
        import tarfile
        import zipfile

        from aiida.common.archive import extract_tar, extract_zip
        from aiida.common.exceptions import NotExistent
        from aiida.common.folders import SandboxFolder
        from aiida.common.json import load as jsonload

        # Get metadata.json and data.json as dicts from v0.1 file archive
        # Cannot use 'get_json_files' for 'export_v0.1_simple.aiida',
        # because we need to pass the SandboxFolder to 'migrate_recursively'
        dirpath_archive = get_archive_file(
            "export_v0.1_simple.aiida", core_file=True)

        with SandboxFolder(sandbox_in_repo=False) as folder:
            if zipfile.is_zipfile(dirpath_archive):
                extract_zip(dirpath_archive, folder, silent=True)
            elif tarfile.is_tarfile(dirpath_archive):
                extract_tar(dirpath_archive, folder, silent=True)
            else:
                raise ValueError(
                    'invalid file format, expected either a zip archive or gzipped tarball'
                )

            try:
                with io.open(
                        folder.get_abs_path('data.json'), 'r',
                        encoding='utf8') as fhandle:
                    data = jsonload(fhandle)
                with io.open(
                        folder.get_abs_path('metadata.json'), 'r',
                        encoding='utf8') as fhandle:
                    metadata = jsonload(fhandle)
            except IOError:
                raise NotExistent(
                    'export archive does not contain the required file {}'.
                    format(fhandle.filename))

            verify_metadata_version(metadata, version='0.1')

            # Migrate to newest version
            new_version = migrate_recursively(metadata, data, folder)
            verify_metadata_version(metadata, version=newest_version)
            self.assertEqual(new_version, newest_version)

    @with_temp_dir
    def test_no_node_export(self, temp_dir):
        """Test migration of export file that has no Nodes"""
        input_file = get_archive_file('export_v0.3_no_Nodes.aiida')
        output_file = os.path.join(temp_dir, 'output_file.aiida')

        # Known entities
        computer_uuids = [self.computer.uuid]  # pylint: disable=no-member
        user_emails = [orm.User.objects.get_default().email]

        # Known export file content used for checks
        node_count = 0
        computer_count = 1 + 1  # localhost is always present
        computer_uuids.append('4f33c6fd-b624-47df-9ffb-a58f05d323af')
        user_emails.append('aiida@localhost')

        # Perform the migration
        migrate_archive(input_file, output_file)

        # Load the migrated file
        import_data(output_file, silent=True)

        # Check known number of entities is present
        self.assertEqual(orm.QueryBuilder().append(orm.Node).count(),
                         node_count)
        self.assertEqual(orm.QueryBuilder().append(orm.Computer).count(),
                         computer_count)

        # Check unique identifiers
        computers = orm.QueryBuilder().append(
            orm.Computer, project=['uuid']).all()[0][0]
        users = orm.QueryBuilder().append(
            orm.User, project=['email']).all()[0][0]
        self.assertIn(computers, computer_uuids)
        self.assertIn(users, user_emails)

    def test_wrong_versions(self):
        """Test correct errors are raised if export files have wrong version numbers"""
        from aiida.cmdline.utils.migration import MIGRATE_FUNCTIONS

        # Initialization
        wrong_versions = ['0.0', '0.12', '0.1.0']
        wrong_versions.append(str(float(newest_version) + 0.1))
        old_versions = list(MIGRATE_FUNCTIONS.keys())
        legal_versions = old_versions + [newest_version]
        wrong_version_metadatas = []
        for version in wrong_versions:
            metadata = {'export_version': version}
            wrong_version_metadatas.append(metadata)

        # Checks
        # Make sure the "wrong_versions" are wrong
        for version in wrong_versions:
            self.assertNotIn(
                version,
                legal_versions,
                msg=
                "'{}' was not expected to be a legal version, legal version: {}"
                .format(version, legal_versions))

        # Make sure migrate_recursively throws a critical message and raises SystemExit
        for metadata in wrong_version_metadatas:
            with self.assertRaises(SystemExit) as exception:
                new_version = migrate_recursively(metadata, {}, None)

                self.assertIn(
                    'Critical: Cannot migrate from version {}'.format(
                        metadata['export_version']),
                    exception.exception,
                    msg=
                    "Expected a critical statement for the wrong export version '{}', "
                    "instead got {}".format(metadata['export_version'],
                                            exception.exception))
                self.assertIsNone(
                    new_version,
                    msg="migrate_recursively should not return anything, "
                    "hence the 'return' should be None, but instead it is {}".
                    format(new_version))

    def test_migrate_newest_version(self):
        """
        Test critical message and SystemExit is raised, when an export file with the newest export version is migrated
        """
        # Initialization
        metadata = {'export_version': newest_version}

        # Check
        with self.assertRaises(SystemExit) as exception:
            new_version = migrate_recursively(metadata, {}, None)

            self.assertIn(
                'Critical: Your export file is already at the newest export version {}'
                .format(metadata['export_version']),
                exception.exception,
                msg=
                "Expected a critical statement that the export version '{}' is the newest export version '{}', "
                "instead got {}".format(metadata['export_version'],
                                        newest_version, exception.exception))
            self.assertIsNone(
                new_version,
                msg="migrate_recursively should not return anything, "
                "hence the 'return' should be None, but instead it is {}".
                format(new_version))

    @with_temp_dir
    def test_v02_to_newest(self, temp_dir):
        """Test migration of exported files from v0.2 to newest export version"""
        # Get export file with export version 0.2
        input_file = get_archive_file('export_v0.2.aiida')
        output_file = os.path.join(temp_dir, 'output_file.aiida')

        # Perform the migration
        migrate_archive(input_file, output_file)
        metadata, _ = get_json_files(output_file)
        verify_metadata_version(metadata, version=newest_version)

        # Load the migrated file
        import_data(output_file, silent=True)

        # Do the necessary checks
        self.assertEqual(orm.QueryBuilder().append(orm.Node).count(),
                         self.node_count)

        # Verify that CalculationNodes have non-empty attribute dictionaries
        builder = orm.QueryBuilder().append(orm.CalculationNode)
        for [calculation] in builder.iterall():
            self.assertIsInstance(calculation.attributes, dict)
            self.assertNotEqual(len(calculation.attributes), 0)

        # Verify that the StructureData nodes maintained their (same) label, cell, and kinds
        builder = orm.QueryBuilder().append(orm.StructureData)
        self.assertEqual(
            builder.count(),
            self.struct_count,
            msg="There should be {} StructureData, instead {} were/was found".
            format(self.struct_count, builder.count()))
        for structures in builder.all():
            structure = structures[0]
            self.assertEqual(structure.label, self.known_struct_label)
            self.assertEqual(structure.cell, self.known_cell)

        builder = orm.QueryBuilder().append(
            orm.StructureData, project=['attributes.kinds'])
        for [kinds] in builder.iterall():
            self.assertEqual(len(kinds), len(self.known_kinds))
            for kind in kinds:
                self.assertIn(
                    kind,
                    self.known_kinds,
                    msg="Kind '{}' not found in: {}".format(
                        kind, self.known_kinds))

        # Check that there is a StructureData that is an input of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.StructureData, tag='structure')
        builder.append(orm.CalculationNode, with_incoming='structure')
        self.assertGreater(len(builder.all()), 0)

        # Check that there is a RemoteData that is the output of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.CalculationNode, tag='parent')
        builder.append(orm.RemoteData, with_incoming='parent')
        self.assertGreater(len(builder.all()), 0)

    @with_temp_dir
    def test_v03_to_newest(self, temp_dir):
        """Test migration of exported files from v0.3 to newest export version"""
        input_file = get_archive_file('export_v0.3.aiida')
        output_file = os.path.join(temp_dir, 'output_file.aiida')

        # Perform the migration
        migrate_archive(input_file, output_file)
        metadata, _ = get_json_files(output_file)
        verify_metadata_version(metadata, version=newest_version)

        # Load the migrated file
        import_data(output_file, silent=True)

        # Do the necessary checks
        self.assertEqual(orm.QueryBuilder().append(orm.Node).count(),
                         self.node_count)

        # Verify that CalculationNodes have non-empty attribute dictionaries
        builder = orm.QueryBuilder().append(orm.CalculationNode)
        for [calculation] in builder.iterall():
            self.assertIsInstance(calculation.attributes, dict)
            self.assertNotEqual(len(calculation.attributes), 0)

        # Verify that the StructureData nodes maintained their (same) label, cell, and kinds
        builder = orm.QueryBuilder().append(orm.StructureData)
        self.assertEqual(
            builder.count(),
            self.struct_count,
            msg="There should be {} StructureData, instead {} were/was found".
            format(self.struct_count, builder.count()))
        for structures in builder.all():
            structure = structures[0]
            self.assertEqual(structure.label, self.known_struct_label)
            self.assertEqual(structure.cell, self.known_cell)

        builder = orm.QueryBuilder().append(
            orm.StructureData, project=['attributes.kinds'])
        for [kinds] in builder.iterall():
            self.assertEqual(len(kinds), len(self.known_kinds))
            for kind in kinds:
                self.assertIn(
                    kind,
                    self.known_kinds,
                    msg="Kind '{}' not found in: {}".format(
                        kind, self.known_kinds))

        # Check that there is a StructureData that is an input of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.StructureData, tag='structure')
        builder.append(orm.CalculationNode, with_incoming='structure')
        self.assertGreater(len(builder.all()), 0)

        # Check that there is a RemoteData that is the output of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.CalculationNode, tag='parent')
        builder.append(orm.RemoteData, with_incoming='parent')
        self.assertGreater(len(builder.all()), 0)



# -*- coding: utf-8 -*-
"""Utility functions to import archives, such as export archives"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import io
import tarfile
import zipfile

from aiida.common.archive import extract_tar, extract_zip
from aiida.common.exceptions import NotExistent
from aiida.common.folders import SandboxFolder
from aiida.common import json


def get_archive_file(archive, core_file=False):
    """
    Return the absolute path of the archive file used for testing purposes. The expected path for these files:

        aiida-export_migration-testing.archives

    :param archive: the relative filename of the archive
    :param core_file: Whether or not the file is located in aiida_core repo or the current repo.
    :returns: absolute filepath of the archive test file
    """
    import inspect

    if core_file:
        # Add filepath to aiida_core repo
        aiida_fixtures_path = os.path.join(
            os.path.dirname(inspect.getabsfile(json)), os.path.pardir,
            "backends/tests/fixtures/export/migrate")

        dirpath_archive = os.path.join(aiida_fixtures_path, archive)
    else:
        # Add filepath to local repo
        dirpath_current = os.path.dirname(os.path.realpath(__file__))
        dirpath_archives = os.path.join(dirpath_current, 'archives')

        dirpath_archive = os.path.join(dirpath_archives, archive)

    if not os.path.isfile(dirpath_archive):
        dirpath_parent = os.path.dirname(dirpath_archive)
        raise ValueError(
            'archive {} does not exist in the archives directory {}'.format(
                archive, dirpath_parent))

    return dirpath_archive


def get_json_files(archive, silent=True, core_file=False):
    """Get metadata.json and data.json from an exported AiiDA archive

    :param archive: the relative filename of the archive
    :param silent: Whether or not the extraction should be silent
    :param core_file: Whether or not the file is located in aiida_core repo or the current repo.
    """
    # Get archive
    dirpath_archive = get_archive_file(archive, core_file=core_file)

    # Unpack archive
    with SandboxFolder(sandbox_in_repo=False) as folder:
        if zipfile.is_zipfile(dirpath_archive):
            extract_zip(dirpath_archive, folder, silent=silent)
        elif tarfile.is_tarfile(dirpath_archive):
            extract_tar(dirpath_archive, folder, silent=silent)
        else:
            raise ValueError(
                'invalid file format, expected either a zip archive or gzipped tarball'
            )

        try:
            with io.open(
                    folder.get_abs_path('data.json'), 'r',
                    encoding='utf8') as fhandle:
                data = json.load(fhandle)
            with io.open(
                    folder.get_abs_path('metadata.json'), 'r',
                    encoding='utf8') as fhandle:
                metadata = json.load(fhandle)
        except IOError:
            raise NotExistent(
                'export archive does not contain the required file {}'.format(
                    fhandle.filename))

    # Return metadata.json and data.json
    return metadata, data


def migrate_archive(input_file, output_file, silent=True):
    """Migrate contents using `migrate_recursively`
    This is essentially similar to `verdi export migrate`.
    However, since this command may be disabled, this function simulates it and keeps the tests working.

    :param input_file: filename with full path for archive to be migrated
    :param output_file: filename with full path for archive to be created after migration
    """
    from aiida.cmdline.utils.migration import migrate_recursively

    # Unpack archive, migrate, and re-pack archive
    with SandboxFolder(sandbox_in_repo=False) as folder:
        if zipfile.is_zipfile(input_file):
            extract_zip(input_file, folder, silent=silent)
        elif tarfile.is_tarfile(input_file):
            extract_tar(input_file, folder, silent=silent)
        else:
            raise ValueError(
                'invalid file format, expected either a zip archive or gzipped tarball'
            )

        try:
            with io.open(
                    folder.get_abs_path('data.json'), 'r',
                    encoding='utf8') as fhandle:
                data = json.load(fhandle)
            with io.open(
                    folder.get_abs_path('metadata.json'), 'r',
                    encoding='utf8') as fhandle:
                metadata = json.load(fhandle)
        except IOError:
            raise NotExistent(
                'export archive does not contain the required file {}'.format(
                    fhandle.filename))

        # Migrate
        migrate_recursively(metadata, data, folder)

        # Write json files
        with io.open(folder.get_abs_path('data.json'), 'wb') as fhandle:
            json.dump(data, fhandle, indent=4)

        with io.open(folder.get_abs_path('metadata.json'), 'wb') as fhandle:
            json.dump(metadata, fhandle, indent=4)

        # Pack archive
        compression = zipfile.ZIP_DEFLATED
        with zipfile.ZipFile(
                output_file, mode='w', compression=compression,
                allowZip64=True) as archive:
            src = folder.abspath
            for dirpath, dirnames, filenames in os.walk(src):
                relpath = os.path.relpath(dirpath, src)
                for filename in dirnames + filenames:
                    real_src = os.path.join(dirpath, filename)
                    real_dest = os.path.join(relpath, filename)
                    archive.write(real_src, real_dest)



# -*- coding: utf-8 -*-
from aiida.orm import DataFactory, CalculationFactory
from aiida.orm.code import Code
from aiida.work.workchain import WorkChain, ToContext, Calc
from aiida.orm.data.upf import get_pseudos_from_structure
from aiida.work.run import submit

UpfData = DataFactory('upf')
ParameterData = DataFactory('parameter')
KpointsData = DataFactory('array.kpoints')
StructureData = DataFactory('structure')
PwCalculation = CalculationFactory('quantumespresso.pw')

class TestWorkChain(WorkChain):

    @classmethod
    def define(cls, spec):
        super(TestWorkChain, cls).define(spec)
        spec.input("code", valid_type=Code)
        spec.input("structure", valid_type=StructureData)
        spec.outline(
            cls.run_scf,
            cls.run_md,
            cls.return_results,
        )
        spec.dynamic_output()

    def run_scf(self):
        structure = self.inputs.structure
        return self._submit_pw_calc(structure, label="scf", runtype='scf',
                                    wallhours=4)
    
    def run_md(self):
        prev_calc = self.ctx.scf
        structure = self.inputs.structure
        return self._submit_pw_calc(structure, label="md", runtype='md',
                                    wallhours=4, parent_folder=prev_calc.out.remote_folder)

    def _submit_pw_calc(self, structure, label, runtype, wallhours=24, parent_folder=None):
        self.report("Running pw.x for "+label)

        inputs = {}
        inputs['_label'] = label
        inputs['code'] = self.inputs.code
        inputs['structure'] = structure
        inputs['parameters'] = self._get_parameters(structure, runtype, parent_folder)
        inputs['pseudo'] = self._get_pseudos(structure,
                                             family_name="SSSP_efficiency_v1.0")
        if parent_folder:
            inputs['parent_folder'] = parent_folder

        # kpoints
        kpoints = KpointsData()
        kpoints_mesh = 2
        kpoints.set_kpoints_mesh([kpoints_mesh, kpoints_mesh, kpoints_mesh])
        inputs['kpoints'] = kpoints

        inputs['_options'] = {
            "resources": {"num_machines": 1, "num_mpiprocs_per_machine": 1},
            "max_wallclock_seconds": wallhours * 60 * 60,  # hours
        }


        future = submit(PwCalculation.process(), **inputs)
	return ToContext(**{label: Calc(future)})

    def _get_parameters(self, structure, runtype, parent_folder=True):
        params = {'CONTROL': {
                     'calculation': runtype,
                     'wf_collect': True,
                     'forc_conv_thr': 0.0001,
                     'nstep': 4,
                     },
                  'SYSTEM': {
                       'ecutwfc': 50.,
                       'ecutrho': 400.,
                       'occupations': 'smearing',
                       'degauss': 0.001,
                       },
                  'ELECTRONS': {
                       'conv_thr': 1.e-8,
                       'mixing_beta': 0.25,
                       'electron_maxstep': 50,
                       'scf_must_converge': False,
                      },
                  }
        if parent_folder:
            params['CONTROL']['restart_mode'] = 'restart'
        return ParameterData(dict=params)

    def _get_pseudos(self, structure, family_name):
        kind_pseudo_dict = get_pseudos_from_structure(structure, family_name)
        pseudos = {}
        for p in kind_pseudo_dict.values():
            ps = [k for k, v in kind_pseudo_dict.items() if v == p]
            kinds = "_".join(ps)
            pseudos[kinds] = p

        return pseudos

    def return_results(self):
        from aiida.orm.data.base import Int, Float
        i = Int()
        f = Float()
        self.out("integer", i)
        self.out("float", f)




# -*- coding: utf-8 -*-

from __future__ import absolute_import
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="aiida-export-migration-tests",
    version="0.1.1",
    author="The AiiDA team",
    author_email="developers@aiida.net",
    description="Export migrations tests for AiiDA",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aiidateam/aiida-export-migration-tests",
    packages=find_packages(),
    license="MIT Licence",
    include_package_data=True)
